package main

import (
	"crypto/rand"
	"flag"
	"fmt"
	"io"
	"log"
	"os"

	"github.com/netfoundry/secretstream"
	"golang.org/x/crypto/argon2"
	"golang.org/x/term"
)

const (
	// default key derivation values in libsodium
	argon2id_time_cost        = 2
	argon2id_memory_cost      = 67108864 / 1024
	argon2id_threads          = 1
	argon2id_salt_len         = 16
	xchacha20poly1305_key_len = 32
	xchacha20poly1305_iv_len  = 16
	xchacha20poly1305_buf_len = 8192
)

var param struct {
	argon2id_time_cost uint64
	argon2id_memory_cost uint64
	argon2id_threads uint64
	argon2id_salt_len int64
	xchacha20poly1305_key_len uint64
	xchacha20poly1305_iv_len uint64
	xchacha20poly1305_buf_len uint64
}

func usage() {
	fmt.Printf("usage: %s [-hed] [-b size] [-j thread] [-t time] [-m memory] [-s salt] [-f file]\n", os.Args[0])
	os.Exit(2)
}

func readsalt(f *os.File, salt *[]byte) {
	_, err := f.Read(*salt)
	if err != nil {
		log.Fatal(err)
	}
}

func deriv(pw []byte, key *[]byte, salt []byte) {
	*key = argon2.IDKey(pw, salt,
		uint32(param.argon2id_time_cost),
		uint32(param.argon2id_memory_cost),
		uint8(param.argon2id_threads),
		uint32(param.xchacha20poly1305_key_len))
}

func encrypt(in *os.File, out *os.File, key []byte, salt []byte) {
	var tag byte
	buf := make([]byte, param.xchacha20poly1305_buf_len)

	enc, nonce, err := secretstream.NewEncryptor(key)
	if err != nil {
		log.Fatal(err)
	}

	out.Write(salt)
	if err != nil {
		log.Fatal(err)
	}

	out.Write(nonce)
	if err != nil {
		log.Fatal(err)
	}

	loop := 1
	tag = secretstream.TagMessage
	for loop > 0 {
		n, err := in.Read(buf)
		if err == io.EOF || n < len(buf) {
			tag = secretstream.TagFinal
			loop = 0
		} else if err != nil {
			log.Fatal(err)
		}
		cipher, err := enc.Push(buf[:n], tag)
		if err != nil {
			log.Fatal(err)
		}

		out.Write(cipher)
		if err != nil {
			log.Fatal(err)
		}
	}
}

func decrypt(in *os.File, out *os.File, key []byte) {
	buf := make([]byte, param.xchacha20poly1305_buf_len+secretstream.StreamABytes)
	header := make([]byte, secretstream.StreamHeaderBytes)

	// Skip beginning of file which (supposedly) contains the salt for the key
	in.Seek(param.argon2id_salt_len, os.SEEK_SET)
	_, err := in.Read(header)
	if err != nil {
		log.Fatal(err)
	}

	dec, err := secretstream.NewDecryptor(key, header)
	if err != nil {
		log.Fatal(err)
	}

	loop := 1
	for loop > 0 {
		n, err := in.Read(buf)
		if err != nil && err != io.EOF {
			log.Fatal(err)
		}

		plain, tag, err := dec.Pull(buf[:n])
		if err != nil {
			log.Fatal(err)
		}

		if tag == secretstream.TagFinal {
			loop = 0
		}
		out.Write(plain)
	}
}

func main() {
	var err error
	var key, salt, pass []byte
	var filename, saltfile string
	var dflag, eflag bool

	in := os.Stdin
	out := os.Stdout

	// Init default cipher values
	param.argon2id_salt_len = argon2id_salt_len
	param.xchacha20poly1305_key_len = xchacha20poly1305_key_len
	param.xchacha20poly1305_iv_len = xchacha20poly1305_iv_len

	salt = make([]byte, param.argon2id_salt_len)
	key = make([]byte, param.xchacha20poly1305_key_len)

	flag.StringVar(&filename, "f", "", "Encrypt/decrypt to/from file name")
	flag.StringVar(&saltfile, "s", "", "Read salt from file (encrypt-only)")
	flag.BoolVar(&eflag, "e", false, "encrypt input (default)")
	flag.BoolVar(&dflag, "d", false, "decrypt input")

	// xchacha20/argon2id parameters
	flag.Uint64Var(&param.argon2id_time_cost, "t", argon2id_time_cost, "Time cost/Iterations")
	flag.Uint64Var(&param.argon2id_memory_cost, "m", argon2id_memory_cost, "Memory cost")
	flag.Uint64Var(&param.argon2id_threads, "j", argon2id_threads, "Parallel threads")
	flag.Uint64Var(&param.xchacha20poly1305_buf_len, "b", xchacha20poly1305_buf_len, "Buffer size")

	flag.Usage = usage
	flag.Parse()

	if eflag && dflag {
		log.Fatal("Cannot use encryption and decryption at the same time")
	}

	args := flag.Args()
	if len(args) > 0 {
		usage()
	}

	// Prompt user for password
	tty, err := os.OpenFile("/dev/tty", os.O_RDWR, 0644)
	if err != nil {
		log.Fatal(err)
	}
	defer tty.Close()
	fmt.Fprint(tty, "password:")
	pass, _ = term.ReadPassword(int(tty.Fd()))
	fmt.Fprintln(tty, "")

	if dflag {
		if len(filename) > 0 {
			in, err = os.Open(filename)
			if err != nil {
				log.Fatal(err)
			}
			defer in.Close()
		}

		readsalt(in, &salt)
		deriv(pass, &key, salt)
		decrypt(in, out, key)
	} else {
		if len(saltfile) > 0 {
			// Read salt from any manually specified file…
			f, err := os.Open(saltfile)
			if err != nil {
				log.Fatal(err)
			}
			readsalt(f, &salt)
			f.Close()
		} else {
			// … or generate a random one
			_, err = rand.Read(salt)
			if err != nil {
				log.Fatal(err)
			}
		}
		deriv(pass, &key, salt)
		if len(filename) > 0 {
			out, err := os.OpenFile(filename, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0600)
			if err != nil {
				log.Fatal(err)
			}
			defer out.Close()
		}
		encrypt(in, out, key, salt)
	}
}
